/////////////////////////////////////////////////////////////////////////////// 
// 
// Pager.cc 
// --------
// Curses based pager interface definition ( second generation ) 
// 
// Design and Implementation by Bjoern Lemke 
// 
// (C)opyright 2025 Bjoern Lemke 
// 
// IMPLEMENTATION MODULE 
// 
// Class: Monitor 
// 
// Description: Cursor based pager implementation 
// 
// Status: CLEAN 
// 
/////////////////////////////////////////////////////////////////////////////// 

#include <stdio.h> 
#include <stdlib.h> 
#include <unistd.h> 

#include "Pager.h" 
#include "Tokenizer.h"
#include "Sleeper.h" 

#define COL_HEADER 1 
#define COL_FOOTER 2 
#define COL_CONTENT 4 
#define COL_SELECTED 5
#define COL_OK 6
#define COL_ERROR 7

#define LEFTOFFSET 1 
#define TOPOFFSET 1 

#define KEY_CNTRL_A 1
#define KEY_CNTRL_D 4
#define KEY_CNTRL_E 5
#define KEY_TAB 9 
#define KEY_RETURN 10 
#define KEY_CNTRL_L 12
#define KEY_CNTRL_T 20
#define KEY_CNTRL_X 24
#define KEY_ESC 27 
#define KEY_CNTRL_6 30
#define KEY_CNTRL_7 31
#define KEY_CNTRL_PLUS 43
#define KEY_CNTRL_MINUS 45
#define KEY_DEL 127


#define CUR_ROW_OFFSET 3 
#define CUR_COL_OFFSET 20 
#define VAL_MAX_LEN 30 
#define VAL_MAX_NUM 20 
#define OK_COL_POS 6 
#define ABORT_COL_POS 16

#define HEADERSIZE 1
#define PAGESIZE 10

#define MSEC_SLEEPDELAY 70
#define MSEC_RESIZEDELAY 500

#define MODE_EDIT "Edit"
#define MODE_PAGE "Page"

#define STATUS_OK "Ok"
#define STATUS_FAILED "Failed"
#define STATUS_NONE "None"

#define HELP_EDIT 1
#define HELP_PAGE 2

Pager::Pager(const Chain& title) : SigHandler() 
{
    
    _footerSize = 5;
    _title = title;
    _helpwin=0;
    
#ifndef LFCNOCURSES

/*  Initialize ncurses  */ 
    initscr(); 
    curs_set(0); 
    start_color(); 
       
    init_pair(COL_HEADER,  COLOR_WHITE,     COLOR_BLUE); 
    init_pair(COL_FOOTER,  COLOR_WHITE,     COLOR_BLUE); 
    init_pair(COL_CONTENT,  COLOR_WHITE,     COLOR_BLACK); 
    init_pair(COL_SELECTED,  COLOR_GREEN,     COLOR_BLACK);
    init_pair(COL_OK,  COLOR_WHITE,     COLOR_BLUE);
    init_pair(COL_ERROR,  COLOR_WHITE,     COLOR_RED);
    
    noecho();
    keypad(stdscr, TRUE); 
    timeout(0); 

    wbkgd(stdscr, COLOR_PAIR(COL_CONTENT));
    
    init(); 

    // we disable SIGINT since this could be used by calling application
    signal(SIGINT, SIG_IGN);
    install(SIGWINCH);
    
#else
    throw Exception(EXLOC, "Curses not supported");
#endif

    // init
    for ( int i=0; i< MAXCMDROW; i++)
	for ( int j=0; j< MAXCMDCOL; j++)
	    _cmdInput[i][j] = 0;

    _curRefresh=0;
    _rowSelected=0;
    _vStart=0;
    _rowIdx=0;
    _offset=0;
    
} 

Pager::~Pager() 
{ 
#ifndef LFCNOCURSES 
// we need endwin and erase to reset to original terminal settings
    endwin();
    erase();
#endif 
} 

void Pager::sigCatch(int sig) 
{ 
#ifndef LFCNOCURSES
    try 
    { 
	endwin(); 
	
	// _rows = LINES;
	// _cols = COLS;
        initscr();
        clear(); 
        refresh(); 

	Sleeper::milliSleep(MSEC_RESIZEDELAY);
	
        install(SIGWINCH);	
    } 
    catch ( Exception e) 
    { 
        Chain msg; 
        e.pop(msg); 
        cout << msg << endl; 
    }
#endif
} 

void Pager::manage()
{
#ifndef LFCNOCURSES
    _rows = LINES;
    _cols = COLS;
	
    _cntrl = 0; 

    _mode=Chain(MODE_EDIT);
    curs_set(1);

    _curCol=0;    
    _curRow=_rows-_footerSize+1;

    _terminate = false;
    _status = Chain(STATUS_OK);
        
    int inputPos=0;
    
    do 
    {	
        switch (_cntrl) 
        { 
	case KEY_CNTRL_X:
	{
	    // _debug=Chain("Cmd=") + getCommand();

	    _mode=Chain(MODE_PAGE);

	    ListT<Chain> schema;		
	    ListT< ListT<Chain> > outData;  
		
	    if ( execute(getCommand(), _msg) )
	    {
		_status = Chain(STATUS_OK);		
		schema = getSchema();		
		outData = getData();	    
		setPager(schema, outData);
		managePager();
	    }
	    else
	    {
		_status = Chain(STATUS_FAILED);
		setPager(schema, outData);		
	    }
	    break;   
	}
	case KEY_TAB:
	{
	    _mode=Chain(MODE_PAGE);
	    managePager();
	    break;
	}
	case KEY_CNTRL_A:
	{
	    _curCol=0;
	    break;
	}
	case KEY_CNTRL_E:
	{
	    int rowPos = _curRow - ( _rows - _footerSize) - 1;	    
	    while ( _cmdInput[rowPos][_curCol] != 0)		
		_curCol++;	    
	    break;
	}
	
	case KEY_LEFT:
	{
	    if ( _curCol > 0 )
		_curCol--;
	    break;
	}
	case KEY_RIGHT :
	{
	    int rowPos = _curRow - ( _rows - _footerSize) - 1;
	    
	    if ( _curCol <  _cols && _cmdInput[rowPos][_curCol] != 0)		
		_curCol++;
	    break;
	}
	case KEY_UP :
	{
	    if ( _curRow > _rows - _footerSize + 1 )
	    {
		_curRow--;
		
		int rowPos = _curRow - ( _rows - _footerSize) - 1;
		
		while ( _cmdInput[rowPos][_curCol] == 0 && _curCol > 0)
		    _curCol--;
	    }
	    break;
	}
	case KEY_DOWN :
	{
	    if ( _curRow <  _rows - 1 )
	    {
		_curRow++;
		
		int rowPos = _curRow - ( _rows - _footerSize) - 1;
		
		_debug=Chain("rowpos=") + Chain(rowPos);
		
		while ( _cmdInput[rowPos][_curCol] == 0 && _curCol > 0)
		    _curCol--;
	    }
	    break;
	}
	case KEY_DEL:
	case KEY_BACKSPACE:
	{
	    if ( _curCol > 0 )
	    {
		_curCol--;
		
		int rowPos = _curRow - ( _rows - _footerSize) - 1;
		
		_debug=Chain("del rowpos=") + Chain(rowPos) + Chain("cur col=") + Chain(_curCol);
		int i=_curCol;
		while ( _cmdInput[rowPos][i] > 0 )
		{
		    _cmdInput[rowPos][i] = _cmdInput[rowPos][i+1];
		    i++;
		}
		
		_debug=Chain("del rowpos=") + Chain(rowPos) + Chain("del=") + Chain(i-_curCol);
		// _cmdInput[rowPos][i] = 0;
	    }
	    else  
	    {
		// search for end of line input of previous line

		int rowPos = _curRow - ( _rows - _footerSize) - 1;


		if ( rowPos > 0 )
		{
		    int i=0;
		    while ( _cmdInput[rowPos-1][i] > 0 )
			i++;
		    
		    _curCol=i;
		    
		    // copy tail of line to end of previous line
		    
		    int j=0;
		    while ( _cmdInput[rowPos][j] > 0 )
		    {
			_cmdInput[rowPos-1][i] = _cmdInput[rowPos][j];
			i++;
			j++;
		    }
		    
		    // shift one line up
		    for ( int i=rowPos; i<MAXCMDROW; i++)
			for ( int j=0; j< MAXCMDCOL; j++)
			    _cmdInput[i][j] = _cmdInput[i+1][j];
		    
		    
		    _curRow--;
		}
	    }
	    break;
	}
	case KEY_RETURN:
	{

	    int rowPos = _curRow - ( _rows - _footerSize) - 1;
	    
	    // shift one line down
	    for ( int i=MAXCMDROW-1; i>rowPos; i--)
		for ( int j=0; j< MAXCMDCOL; j++)
		    _cmdInput[i][j] = _cmdInput[i-1][j];

	    // copy tail to next row
	    for ( int j=_curCol; j< MAXCMDCOL; j++)
	    {
		_cmdInput[rowPos+1][j-_curCol] = _cmdInput[rowPos][j];
	    }

	    // delete tail of current row
	    for ( int j=_curCol; j< MAXCMDCOL; j++)
	    {
		_cmdInput[rowPos][j] = 0;
	    }
	    _curCol=0;
	    _curRow++;

	    break;
	}
	case KEY_CNTRL_PLUS :
	{
	    // increase footer size here
	    _debug=Chain("Detected cntl-+");
	    _footerSize++;
	    _curRow--;
	    break;
	}
	case KEY_CNTRL_MINUS :
	{
	    // shrink footer size here
	    _debug=Chain("Detected cntl--");
	    _footerSize--;
	    _curRow++;
	    break;
	}
	case KEY_RESIZE:
	{
	    _rows = LINES;
	    _cols = COLS;
	    _curRow=_rows-_footerSize+ 1;
	    _curCol=0;
	    break;  
	}
	case KEY_CNTRL_T:
	{
	    _terminate=true;
	    break;
	}	
	case KEY_CNTRL_7:
	{
	    showHelp(HELP_EDIT);
	    break;
	}	
	case KEY_CNTRL_6:
	{
	    showMessage();
	    break;
	}	
	case KEY_CNTRL_D:
	{
	    _curRow = _rows - _footerSize + 1;
	    _curCol=0;
	    // init
	    for ( int i=0; i< MAXCMDROW; i++)
		for ( int j=0; j< MAXCMDCOL; j++)
		    _cmdInput[i][j] = 0;
	    break;
	}	
	default:
	{
	    if ( isCmdInput(_cntrl) )
	    {
		int rowPos = _curRow - ( _rows - _footerSize) - 1;	      
		
		// mvprintw(, col, "%s", (char*)_cntrl);

		// shift tail
	        int i = _curCol;
		while ( _cmdInput[rowPos][i] != 0 )
		    i++;

		_debug=Chain("end at=") + Chain(i);
		
		while ( i > _curCol )
		{
		    _cmdInput[rowPos][i] = _cmdInput[rowPos][i-1];
		    i--;
		}

		if ( _cmdInput[rowPos][_curCol] == 0 )
		    _cmdInput[rowPos][_curCol+1] = 0;
		
		_cmdInput[rowPos][_curCol] = _cntrl;
				    
		_curCol++;
		// inputPos++;
	    }
	    break;
	}	
	} 

	// curs_set(1);
	    
	if ( ! _terminate )
	{
	    writePage();
	    
	    if ( _status == Chain(STATUS_FAILED) )
	    {
		showMessage();		
		_status=Chain(STATUS_NONE);	    
		wrefresh(stdscr);
		_cntrl = 0; 
	    }
	    else
	    {	    
		wtimeout(stdscr, MSEC_SLEEPDELAY); 
		_cntrl = wgetch(stdscr);
	    }
	}
	
    } 
    while( ! _terminate );             // continue until user hits q 
        
    // refresh(); 
    // clear();
    // endwin();

#endif 
    return; 
}

void Pager::setPager( const ListT<Chain>& schema, const ListT< ListT<Chain> >& outData ) 
{
#ifndef LFCNOCURSES

    _schema = schema;
    
    // init outLen list
    _outLen.Empty();
    Chain *pAttr = _schema.First();
    while ( pAttr )
    {
	Tokenizer t1(*pAttr, ":");
	Chain a;
	t1.nextToken(a);
	_outLen.Insert(a.length() + 1);
	pAttr = _schema.Next();
    }

    formatData(outData);

    _curRefresh=0;
    _rowSelected=0;
    _vStart=0;
    _rowIdx=0;
    _offset=0;

}

void Pager::managePager()
{
    curs_set(0);
    
    _rows = LINES;
    _cols = COLS;

    
    int i, ch = 0; 
    
    do 
    { 	
        switch (ch) 
        { 
	case KEY_DOWN :
	{
	    nextRow();
	    break;	    
	}
	case KEY_UP :
	{
	    prevRow();
	    break;
	}
	case 's' :
	{
	    firstRow();
	    break;
	}
	case 'e' :
	{
	    lastRow();
	    break;
	}
	case 'n' : 
	case ' ' :
	case KEY_NPAGE :
	{
	    nextRow();
	    int pdi=_pdi;
	    int pci=_pci;
	    for ( int i=0; i < PAGESIZE  - 1; i++ )
		nextRow();
	    _pdi=pdi;
	    _pci=pci;
	    break;
	}
	case 'p' :
	case KEY_PPAGE :
	{
	    prevRow();	    
	    int pdi=_pdi;
	    int pci=_pci;

	    for ( int i=0; i < PAGESIZE - 1; i++ )
		prevRow();
	    _pdi=pdi;
	    _pci=pci;
	    
	    break;
	}
	case KEY_RESIZE:
	{
	    _rows = LINES;
	    _cols = COLS;
	    _curRow=_rows-_footerSize+ 1;
	    break;  
	}
	case KEY_CNTRL_T:
	{
	    _terminate=true;
	    break;
	}	
	case KEY_CNTRL_7:
	{
	    showHelp(HELP_PAGE);
	    break;
	}	
	default:
	{
	    break;
	}	
	} 

	curs_set(0);
	
	writePage();
		
	wtimeout(stdscr, MSEC_SLEEPDELAY); 
	ch = wgetch(stdscr);
	
    } 
    while( ch != KEY_TAB && ! _terminate );             // continue until user hits q 

    _mode=Chain(MODE_EDIT);
    curs_set(1);
    
    // refresh(); 
    // clear();
    // endwin();

#else 
    cout << "Curses not enabled" << endl; 
#endif 
    return; 
}

void Pager::nextRow()
{
    
    if ( _rowIdx < _rowHeight.Size()  - 1 )
    {
	if ( _rowSelected >= _rows + _vStart - _rowHeight[_rowIdx] - HEADERSIZE - _footerSize)
	    _vStart += _rowHeight[_rowIdx];
	else
	    _curRefresh=1;
	
	_rowSelected += _rowHeight[_rowIdx];

	_pdi = _rowSelected - _rowHeight[_rowIdx];
	_pci = _rowSelected - _rowHeight[_rowIdx] - _vStart + HEADERSIZE;
    
	_ndi = _rowSelected;
	_nci = _rowSelected - _vStart + HEADERSIZE;

	_rowIdx++;
	
    }
    else
    {
	// if end of outData not reached, scroll forward linewise and remember offset
	if ( _vStart < _outData.Size() - _rows + ( HEADERSIZE + _footerSize ) )
	{
	    _vStart += 1;
	    _offset++;
	}
    }
}

void Pager::prevRow()
{

    if ( _rowIdx > 0 )
    {
	_rowIdx--;
	
	if (_vStart >= _rowSelected - _rowHeight[_rowIdx] && _vStart > 0)
	{
	    _rowSelected -= _rowHeight[_rowIdx];
	    _vStart = _rowSelected;
	    
	    // _debug = Chain("page mode up, ") + Chain(" RowSelectd=") + Chain(_rowSelected) + Chain(" vstart=") + Chain(_vStart);
	}
	else
	{
	    _curRefresh=1;
	    _rowSelected -= _rowHeight[_rowIdx];
	    // _debug = Chain("cursor mode up, ") + Chain(" RowSelectd=") + Chain(_rowSelected) + Chain(" vstart=") + Chain(_vStart);

	    _pdi = _rowSelected + _rowHeight[_rowIdx];
	    _pci = _rowSelected-_vStart + _rowHeight[_rowIdx] + 1;
	    
	    _ndi = _rowSelected;
	    _nci = _rowSelected - _vStart + HEADERSIZE;
	}
    }
}

void Pager::firstRow()
{    
    _rowIdx = 0;    
    _vStart = 0;
    _rowSelected = 0;
}

void Pager::lastRow()
{    
    _rowIdx = _rowHeight.Size() - 1;    
    _vStart = _outData.Size() - _rows +  HEADERSIZE;
    _rowSelected = _outData.Size() - _rowHeight[_rowIdx] - _footerSize;
}

void Pager::writeHeader() 
{ 
#ifndef LFCNOCURSES 

    attron(A_BOLD); 
    wcolor_set(stdscr, COL_HEADER, NULL);

    // clear line
    for(int i = 0; i < _cols; i++) 
    {
	mvprintw(0, i, "%s", (char*)" "); 
    } 

    move(0,0); 

    int col=0;
    Chain *pAttr = _schema.First();
    int colPos=0;
    while ( pAttr ) 
    {
	Chain attr;
	int len;
        getAttrSpec(colPos, *pAttr, attr, len); 
        mvprintw(0, col, "%s", (char*)attr); 

	colPos++;
        col += len;
	pAttr = _schema.Next();
    } 

    attroff(A_BOLD); 
#else 
    throw Exception(EXLOC, "Curses not supported"); 
#endif 

} 

void Pager::writePage()
{

    if ( _curRefresh == 1 )
    {
	refreshRow(_pdi, _pci);
	refreshRow(_ndi, _nci);	
	_curRefresh=0;
	return;
    }

    writeHeader(); 

    
    wcolor_set(stdscr, COL_CONTENT, NULL);

    for(int i = 0; (i< _rows - 1 -_footerSize) && (i + _vStart < _outData.Size()); i++) 
    { 
	for(int j = 0; j < _cols; j++) 
	{
	    mvprintw(i + HEADERSIZE, j, "%s", (char*)" "); 
	} 
	
	int col=0;
	Chain *pAttr = _schema.First();
	int colPos=0;
	while ( pAttr ) 
	{
	    Chain attr;
	    int len;
	    getAttrSpec(colPos, *pAttr, attr, len); 
	    
	    char* pVal = (char*)(_outData[i+_vStart][colPos]);
	    if ( pVal )
		mvprintw(i + HEADERSIZE, col, "%s", pVal); 
	    col += len;
	    colPos++;
	    pAttr = _schema.Next();
	} 	    
    }

    for(int i = _outData.Size(); (i< _rows - 1 -_footerSize); i++) 
    { 
	for(int j = 0; j < _cols; j++) 
	{
	    mvprintw(i + HEADERSIZE, j, "%s", (char*)" "); 
	} 
    }

    
    if ( _rowSelected-_vStart+1 >= 0 && _mode == Chain(MODE_PAGE))
	refreshRow(_rowSelected, _rowSelected-_vStart+1);


    // writeDebug(); 

   
    writeFooter();

    refresh();
        
}


void Pager::writeFooter()
{
#ifndef LFCNOCURSES
    
    wcolor_set(stdscr, COL_FOOTER, NULL);

    attron(A_REVERSE);
    attron(A_BOLD);
    for(int i = 0; i < _cols; i++)
    {
        mvprintw(_rows-_footerSize, i, "%s", (char*)" ");
    }
    
    mvprintw(_rows-_footerSize, 0, " %s", (char*)_mode);

    if ( _mode == Chain(MODE_PAGE) )
    {
	if ( _status != Chain(STATUS_NONE) )
	{
	    if ( _status == Chain(STATUS_OK) )	 
		wcolor_set(stdscr, COL_OK, NULL);
	    else
	    {
		wcolor_set(stdscr, COL_ERROR, NULL);
	    }    
	    mvprintw(_rows-_footerSize, 7, " %s", (char*)_status);
	}
    }
    
    wcolor_set(stdscr, COL_FOOTER, NULL);
    mvprintw(_rows-_footerSize, 30, "( Ctl-7 for Help)");
	     
    // mvprintw(_rows-_footerSize, 0, "Cntrl =  %d          ",_cntrl);

    mvprintw(_rows-_footerSize,_cols-23,"Row %04d ( %04d total)", _rowIdx+1, _rowHeight.Size());
    attroff(A_BOLD);
    attroff(A_REVERSE);

    wcolor_set(stdscr, COL_CONTENT, NULL);

    for ( int i=1; i<_footerSize; i++)
    {
	for(int j = 0; j < _cols; j++)
	{
	    mvprintw(_rows-_footerSize + i, j, "%s", (char*)" ");
	}
    }
    
    for ( int i=0; i<_footerSize; i++)
    {
	mvprintw(_rows-_footerSize+i+1, 0, "%s", _cmdInput[i]);
    }
    wmove(stdscr, _curRow, _curCol);


#endif
}

void Pager::writeDebug() 
{ 
#ifndef LFCNOCURSES 

    wcolor_set(stdscr, COL_FOOTER, NULL);
    
    attron(A_REVERSE);
    move(_rows-1,0); 
    for(int i = 0; i < _cols; i++) 
    { 
        addch(' '); 
    } 

    mvprintw(_rows-1,0, "%s", (char*)_debug); 

    attroff(A_REVERSE); 
#endif 
} 


void Pager::refreshRow(int dataRow, int pageRow)
{
    if ( _outData.Size() == 0)
	return;
    
    int colno=0;
    
    // int rowColor = COL_CONTENT;

    wcolor_set(stdscr, COL_CONTENT, NULL);

    if ( _rowSelected == dataRow )
    {
	wcolor_set(stdscr, COL_SELECTED, NULL);
	wattron(stdscr, A_REVERSE);
	// rowColor = COL_SELECTED;
    }
    
    for(int i = 0; i < _cols; i++) 
    { 
        mvwprintw(stdscr, pageRow, i, "%s", (char*)" ");
    } 

    Chain *pAttr = _schema.First();
    int colPos=0;
    Chain *pVal = _outData[dataRow].First();
    while ( pAttr && pVal )
    {		       
	Chain attr;
	int len;
	getAttrSpec(colPos, *pAttr, attr, len);
		    
	if ( 1 ) // perhaps later we also make right align : lenInfo[0] == 'l' )
	{
	    if ( *pVal )
		mvwprintw(stdscr, pageRow, colno, "%s", (char*)*pVal);
	}
	else
	{
	    int offset = len - pVal->length();
	    if ( *pVal )
		mvwprintw(stdscr, pageRow, colno+offset, "%s", (char*)*pVal);	    
	}

	colPos++;
	colno = colno + len;
	
	// wcolor_set(stdscr, rowColor, NULL);
	
	pVal = _outData[dataRow].Next();
	
	pAttr = _schema.Next();
    }
    
    
    if ( _rowSelected == dataRow )
    {		
	// wcolor_set(stdscr, COL_CONTENT, NULL);
	wattroff(stdscr, A_REVERSE); 		
    }
}

void Pager::getAttrSpec(int colPos, const Chain& s, Chain& attr, int& len)
{    
    Tokenizer t1(s, ":");
    t1.nextToken(attr);
    Chain attrLen;
    t1.nextToken(attrLen);
    if ( attrLen == Chain("auto") )
	len = _outLen[colPos];
    else
	len = attrLen.asInteger();
}

void Pager::formatData(const ListT< ListT<Chain> > & outData)
{

    _outData.Empty();
    _rowHeight.Empty();
    
    ListT<Chain> *pRow = outData.First();
    
    int rowNo=0;
    while ( pRow )
    {
	int lineNo=0;
	
	bool moreLines=true;
	
	while ( moreLines )
	{
	    ListT<Chain> outRow;
	    
	    moreLines=false;

	    int colNum=0;
	    Chain *pCol = pRow->First();
	    while ( pCol )
	    {
		Tokenizer tok(*pCol, "\n");

		int i=0;
		Chain s;
		bool hasToken=false;
		while (  i <= lineNo )
		{
		    hasToken=tok.nextToken(s);		    		    
		    i++;
		}
		if ( hasToken )
		{
		    // cout << "Got token " << s << " for lineno = " << lineNo << endl;
		    outRow.Insert(s);

		    if ( _outLen[colNum] < s.length() + 1 )
			_outLen[colNum] = s.length() + 1;
		    
		    moreLines=true;
		}
		else
		{
		    // cout << "Got empty for lineno = " << lineNo << endl;
		    Chain empty;
		    outRow.Insert(empty);
		}
		colNum++;
		pCol = pRow->Next();
	    }
	    
	    if ( moreLines )
	    {
		_outData.Insert(outRow);
		lineNo++;
		rowNo++;
	    }
	}

	// cout << "Insert line height" << lineNo << endl;
	_rowHeight.Insert(lineNo);
	
	pRow = outData.Next();
    }
}

bool Pager::isCmdInput(int c)
{
    if ( isalnum (c)
	 || c == '_'
	 || c == '|'
	 || c == '-'
	 || c == '\''
	 || c == '/'
	 || c == '.'
	 || c == ','
	 || c == ';'
	 || c == '*'
	 || c == '['
	 || c == ']'
	 || c == ':'
	 || c == '@'
	 || c == '='
	 || c == '!'
	 || c == '?'
	 || c == '*'
	 || c == '+'
	 || c == '#'
	 || c == '('
	 || c == ')'
	 || c == '{'
	 || c == '}'
	 || c == '['
	 || c == ']'
	 || c == '%'
	 || c == '&'
	 || c == '$'
	 || c == '>'
	 || c == '<'
	 || c == ' ' )
	return true;
    return false;
}

Chain Pager::getCommand()
{
    Chain cmd;
    
    for ( int i=0; i< MAXCMDROW; i++)
    {
	Chain chunk(_cmdInput[i]);
	cmd = cmd + Chain(" ") + chunk;
    }
    return cmd.cutTrailing(Chain(" "));
}


void Pager::showHelp(int helpMode)
{

    curs_set(0);
    
    int startx, starty;

    int xoff,yoff;
    getbegyx(stdscr, yoff, xoff);
    
    int x,y;
    getmaxyx(stdscr, y, x);

    startx = xoff + (x / 7);
    starty = yoff + (y / 7);

    
    WINDOW* helpwin = newwin( 20, 50, starty, startx);

    // wclear(_helpwin);

    keypad(helpwin, FALSE);
  
    wattron(helpwin, A_BOLD);

    if ( helpMode == HELP_PAGE )
	mvwprintw(helpwin, 1, 2, "%s", (char*)"Page Commands");
    else
	mvwprintw(helpwin, 1, 2, "%s", (char*)"Edit Commands");
    
    wattroff(helpwin, A_BOLD);

    box(helpwin, 0, 0);

    int i=3;
    if ( helpMode == HELP_PAGE )
    {
	mvwprintw(helpwin, i++, 2, "%s",  (char*)"Up/Down - Next / Previous line");
	mvwprintw(helpwin, i++, 2, "%s",  (char*)"s       - Skip to first line");
	mvwprintw(helpwin, i++, 2, "%s",  (char*)"e       - Skip to last line");
	mvwprintw(helpwin, i++, 2, "%s",  (char*)"n       - Next page");
	mvwprintw(helpwin, i++, 2, "%s",  (char*)"p       - Previous page");
	mvwprintw(helpwin, i++, 2, "%s",  (char*)"Cntrl-7 - Show help");       
	mvwprintw(helpwin, i++, 2, "%s",  (char*)"Tab     - Switch to edit mode");
	mvwprintw(helpwin, i++, 2, "%s",  (char*)"Cntrl-t - Terminate");
    }
    else
    {
	mvwprintw(helpwin, i++, 2, "%s",  (char*)"Cursor Key - Navigate");
	mvwprintw(helpwin, i++, 2, "%s",  (char*)"Cntrl-a    - Start of line");
	mvwprintw(helpwin, i++, 2, "%s",  (char*)"Cntrl-e    - End of line");
	mvwprintw(helpwin, i++, 2, "%s",  (char*)"Cntrl-d    - Clear");
	mvwprintw(helpwin, i++, 2, "%s",  (char*)"Cntrl-x    - Execute command");
	mvwprintw(helpwin, i++, 2, "%s",  (char*)"Cntrl-6    - Show last message");
	mvwprintw(helpwin, i++, 2, "%s",  (char*)"Cntrl-7    - Show help");       
	mvwprintw(helpwin, i++, 2, "%s",  (char*)"Cntrl-+/-  - Increase / Decrease edit area");
	mvwprintw(helpwin, i++, 2, "%s",  (char*)"Tab        - Switch to page mode");
	mvwprintw(helpwin, i++, 2, "%s",  (char*)"Cntrl-t    - Terminate");
    }
    
    int c = wgetch(helpwin);

    wrefresh(helpwin);
    
    delwin(helpwin);
    refresh();
    wclear(stdscr);

    curs_set(1);
    
}


void Pager::showMessage()
{

    curs_set(0);
    
    int startx, starty;

    int xoff,yoff;
    getbegyx(stdscr, yoff, xoff);
    
    int x,y;
    getmaxyx(stdscr, y, x);

    startx = xoff + (x / 7);
    starty = yoff + (y / 4);

    
    WINDOW* msgwin = newwin( 5, _cols-10, starty, 5);

    keypad(msgwin, FALSE);
  
    wattron(msgwin, A_BOLD);
    mvwprintw(msgwin, 1, 2, "%s", (char*)"Query Message");
    wattroff(msgwin, A_BOLD);
    box(msgwin, 0, 0);

    mvwprintw(msgwin, 1, 2, "%s", (char*)_msg);

    int c = wgetch(msgwin);

    wclear(msgwin);
    delwin(msgwin);
    wrefresh(stdscr);
    
    curs_set(1);
    
}
